/*******************************************************************************
 * Copyright (c) 2011 INRIA.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Cau Clasen - initial API and implementation
 *******************************************************************************/

package fr.inria.virtualemf.core;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;

import fr.inria.virtualemf.elements.MergeElementImpl;
import fr.inria.virtualemf.elements.ReproduceElementImpl;

public class MetamodelManager {

	private Resource compositionMetamodel;
	private List<Resource> contributingMetamodels = new ArrayList<Resource>();

	private Map<String, EClass> mergeClassesByName = new HashMap<String, EClass>();

	private Map<EStructuralFeature, EStructuralFeature> virtualToConcreteFeature = new HashMap<EStructuralFeature, EStructuralFeature>();
	private Map<EStructuralFeature, List<EStructuralFeature>> mergedToConcreteFeatures = new HashMap<EStructuralFeature, List<EStructuralFeature>>();
	private Map<EStructuralFeature, EStructuralFeature> concreteToVirtualFeature = new HashMap<EStructuralFeature, EStructuralFeature>();
	private Map<EStructuralFeature, EStructuralFeature> concreteToMergedFeature = new HashMap<EStructuralFeature, EStructuralFeature>();
	private Map<String, List<EStructuralFeature>> virtualAssociations = new HashMap<String, List<EStructuralFeature>>();

	private Map<EClass, EClass> concreteToVirtualClass = new HashMap<EClass, EClass>();
	private Map<EClass, List<EClass>> mergeToConcreteClasses = new HashMap<EClass, List<EClass>>();

	public MetamodelManager(List<Resource> contributingMetamodels, Resource compositionMetamodel, VirtualModel virtualModel){
		this.compositionMetamodel = compositionMetamodel;
		this.contributingMetamodels = contributingMetamodels;
		for (Resource metamodel : contributingMetamodels) {
			VirtualModelUtil.registerMetamodel(metamodel, virtualModel.getVirtualResourceSet());
		}
		this.buildMaps();
	}

	private void buildMaps() {
		Map<String, List<EClass>> contributingClassesByName = new HashMap<String, List<EClass>>();
		Map<String, List<EClass>> compositionClassesByName = new HashMap<String, List<EClass>>(); //excluding merged classes

		//builds composition ECLasses by name map
		//builds merged ECLasses by name map
		for(Iterator<EObject> i = compositionMetamodel.getAllContents(); i.hasNext();){
			EObject obj = i.next();
			if (obj instanceof EClass) {
				if(((EClass) obj).getEPackage().getName().equals("MergePackage")){
					mergeClassesByName.put(((EClass) obj).getName(), (EClass) obj);
				} else {
					if(compositionClassesByName.get(((EClass)obj).getName()) == null){
						List<EClass> ecs = new ArrayList<EClass>();
						ecs.add((EClass) obj);
						compositionClassesByName.put(((EClass)obj).getName(), ecs);
					} else {
						compositionClassesByName.get(((EClass) obj).getName()).add((EClass) obj);
					}
				}
			}
		}

		// builds contributing classes by name map
		// registers contributing EPackages
		for(Resource metamodel : contributingMetamodels){
			for(Iterator<EObject> i = metamodel.getAllContents(); i.hasNext();) {
				EObject obj = i.next();
				if(obj instanceof EClass) {
					if(contributingClassesByName.get(((EClass)obj).getName()) == null){
						List<EClass> ecs = new ArrayList<EClass>();
						ecs.add((EClass) obj);
						contributingClassesByName.put(((EClass)obj).getName(), ecs);
					} else {
						contributingClassesByName.get(((EClass) obj).getName()).add((EClass) obj);
					}
				}
			}
		}

		// builds the maps relating virtual and contributing classes and features for 1-to-1 virtual elements
		for(List<EClass> lcec : contributingClassesByName.values()){
			if (lcec.size() > 1){
				for (EClass cec : lcec) {
					List<EClass> lvec = compositionClassesByName.get(cec.getName());
					for (EClass vec : lvec) {
						if(vec.getEPackage().getNsURI().equals("_virtual_" + cec.getEPackage().getNsURI())){
							this.concreteToVirtualClass.put(cec, vec);
							mapFeatures(cec, vec);
						}
					}
				}
			} else {
				EClass cec = lcec.get(0);
				EClass vec = compositionClassesByName.get(cec.getName()).get(0);
				this.concreteToVirtualClass.put(cec, vec);
				mapFeatures(cec, vec);
			}
		}

		// builds the maps relating merged and contributing classes and features for merged elements
//		for(EClass mec : mergeClassesByName.values()){
//			for(EClass cec : mergedToConcreteClasses(mec)){
//				if(mergeToConcreteClasses.get(mec) == null){
//					List<EClass> concClasses = new ArrayList<EClass>();
//					concClasses.add(cec);
//					mergeToConcreteClasses.put(mec, concClasses);
//				} else {
//					mergeToConcreteClasses.get(mec).add(cec);
//				}
//				for (EStructuralFeature csf : cec.getEAllStructuralFeatures()) {
//					EStructuralFeature msf = mec.getEStructuralFeature(csf.getName());
//					if(mergedToConcreteFeatures.get(msf) == null){
//						List<EStructuralFeature> sfs = new ArrayList<EStructuralFeature>();
//						sfs.add(csf);
//						mergedToConcreteFeatures.put(msf, sfs);
//					} else {
//						mergedToConcreteFeatures.get(msf).add(csf);
//					}
//					this.concreteToMergedFeature.put(csf, msf);
//				}
//			}
//		}

		// builds the list of virtual associations (we suppose virtual associations cannot be created for merged elements)
		for (List<EClass> lec : compositionClassesByName.values()) {
			for (EClass ec : lec) {
				for (EStructuralFeature sf : ec.getEStructuralFeatures()) {
					if(virtualToConcreteFeature.get(sf) == null)
						if(virtualAssociations.get(sf.getName()) == null){
							List<EStructuralFeature> sfs = new ArrayList<EStructuralFeature>();
							sfs.add(sf);
							virtualAssociations.put(sf.getName(), sfs);
						} else {
							virtualAssociations.get(sf.getName()).add(sf);
						}
				}
			}
		}
	}

	private List<EClass> mergedToConcreteClasses(EClass mec){
		List<EClass> matchedClasses = new ArrayList<EClass>();
		for(EClass cec : concreteToVirtualClass.keySet()){
			if(compareClasses(mec, cec)){
				matchedClasses.add(cec);
			}
		}
		//remove superTypes
		List<EClass> toRemove = new ArrayList<EClass>();;
		for (EClass ec : matchedClasses) {
			List<EClass> superTypes = ec.getEAllSuperTypes();
			for (EClass st : superTypes) {
				if(matchedClasses.contains(st)){
					toRemove.add(st);
				}
			}
		}
		for (EClass ec : toRemove) {
			matchedClasses.remove(ec);
		}
		//just another checking to be sure...
		int expected = mec.getEAllStructuralFeatures().size();
		int nbSFs = 0;
		for (EClass eClass : matchedClasses) {
			nbSFs += eClass.getEAllStructuralFeatures().size();
		}
		if(expected != (nbSFs - searchForDuplicatedFeatures(matchedClasses).size())){
			throw new VirtualModelOperationException("wrong matching of feature numbers...");
		}
		return matchedClasses;
	}

	private void mapFeatures(EClass concEC, EClass virtuEC) {
		for(EStructuralFeature feature : concEC.getEStructuralFeatures()) {
			EStructuralFeature vf = virtuEC.getEStructuralFeature(feature.getName());
			this.virtualToConcreteFeature.put(vf, feature);
			this.concreteToVirtualFeature.put(feature, vf);
		}
	}

	//checks if cec is not equivalent to a supertype of mec
	private boolean isSuperType(EClass mec, EClass cec){
		for (EClass st : mec.getEAllSuperTypes()) {
			if(cec.getName().equals(st.getName()) && cec.getEAllStructuralFeatures().size() == st.getEAllStructuralFeatures().size())
				return true;
		}
		return false;
	}

	// checks if all features from concrete are present in merged
	private boolean compareClasses(EClass mec, EClass cec){
		if(isSuperType(mec, cec)){
			return false;
		}
		List<EStructuralFeature> matchedFeatures = new ArrayList<EStructuralFeature>();
		for(EStructuralFeature csf : cec.getEStructuralFeatures()){
			for(EStructuralFeature msf : mec.getEStructuralFeatures()){
				if(VirtualModelUtil.compareFeatures(csf, msf)){
					matchedFeatures.add(csf);
				}
			}
		}
		return cec.getEStructuralFeatures().size() == matchedFeatures.size();
	}

	public EStructuralFeature translateToVirtualFeature(EObject object, EStructuralFeature feature) {
		if(object instanceof ReproduceElementImpl)
			return this.concreteToVirtualFeature.get(feature);
		else if (object instanceof MergeElementImpl){
			if (this.concreteToMergedFeature.get(feature) == null){
				System.out.println();
				return this.concreteToVirtualFeature.get(feature);
			} else {
				return this.concreteToMergedFeature.get(feature);
			}
		}
		return null;
	}

	public EStructuralFeature translateFromVirtualFeature(EObject object, EStructuralFeature feature) {
		if(object instanceof ReproduceElementImpl)
			return this.virtualToConcreteFeature.get(feature);
		else if (object instanceof MergeElementImpl)
			throw new VirtualModelOperationException("could not translate feature for merged element");
		return null;
	}

	public List<EStructuralFeature> translateFromMergedFeature(EObject object, EStructuralFeature feature) {
//		return mergedToConcreteFeatures.get(feature);
		List<EStructuralFeature> csfs = new ArrayList<EStructuralFeature>();
		List<EObject> ceos = ((MergeElementImpl)object).getConcreteElements();
		for (EObject eo : ceos) {
			EStructuralFeature sf = eo.eClass().getEStructuralFeature(feature.getName());
			if(sf != null){
				csfs.add(sf);
			}
		}
		return csfs;
	}

	public EClass translateToVirtualEClass(EClass eClass) {
		return this.concreteToVirtualClass.get(eClass);
	}

	public EClass getMergeClass(String mergeClassName){
		return mergeClassesByName.get(mergeClassName);
	}
	
//	public EClass getMergeClass(List<EObject> concObjs){
//		for (EClass mec : mergeToConcreteClasses.keySet()) {
//			int matches = 0;
//			for (EObject eo : concObjs) {
//				if(mergeToConcreteClasses.get(mec).contains(eo.eClass())){
//					matches++;
//				}
//			}
//			if(matches == concObjs.size()){
//				return mec;
//			}
//		}
//		throw new VirtualModelOperationException("Could not find Merge class");
//	}

	public List<EStructuralFeature> searchForDuplicatedFeatures(List<EClass> eClasses){
		if(eClasses.size() > 2)
			throw new VirtualModelOperationException("Don`t support the mergind of more than two classes for the time being...");
		List<EStructuralFeature> dups = new ArrayList<EStructuralFeature>();
		for (EStructuralFeature leftSF : eClasses.get(0).getEAllStructuralFeatures()) {
			for (EStructuralFeature rightSF : eClasses.get(1).getEAllStructuralFeatures()) {
				if(VirtualModelUtil.compareFeatures(leftSF, rightSF)){
					dups.add(leftSF);
				}
			}
		}
		return dups;
	}

	public EStructuralFeature getVirtualAssociation(EObject object, String name){
		List<EStructuralFeature> sfs = virtualAssociations.get(name);
		for (EStructuralFeature sf : sfs) {
			if(object.eClass().getEAllStructuralFeatures().contains(sf)){
				return sf;
			}
		}
		throw new VirtualModelOperationException("could not find virtual association");
	}

	public boolean isVirtualAssociation(EStructuralFeature feature) {
		List<EStructuralFeature> sfs = virtualAssociations.get(feature.getName());
		if(sfs == null)
			return false;
		else
			return sfs.contains(feature);
	}
}
