/*******************************************************************************
 * Copyright (c) 2011 INRIA.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Cau Clasen - initial API and implementation
 *******************************************************************************/

package fr.inria.virtualemf.core;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EPackage.Registry;
import org.eclipse.emf.ecore.impl.EPackageRegistryImpl;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceImpl;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.BasicExtendedMetaData.EPackageExtendedMetaData;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl;


public class VirtualModel extends ResourceImpl {

	private Properties virtualModelProperties = new Properties();
	private List<Resource> contributingModels;
	
	private VirtualLinkManager vLinkManager;
	
	private EList<EObject> virtualContents;
	
	private ResourceSet virtualResourceSet;
	
	private MetamodelManager metamodelManager;
	
	public VirtualModel(URI uri) {
		super(uri);
	}
	
	@SuppressWarnings("unchecked")
	@Override
	protected void doLoad(InputStream inputStream, Map<?, ?> options) throws IOException {
		
		// this is an internal resourceSet only for the virtualized models (to hide them for external applications)
		virtualResourceSet = new ResourceSetImpl();
		
		virtualModelProperties.load(inputStream);
				
		//necessary for when the virtual resource is loaded dynamically and the global factory registry is not set (it adds the factories registered when creating the virtual model)
		Resource.Factory.Registry.INSTANCE.getExtensionToFactoryMap().putAll(this.getResourceSet().getResourceFactoryRegistry().getExtensionToFactoryMap());
		// force loading of UML models as XMIResource
		Resource.Factory.Registry.INSTANCE.getExtensionToFactoryMap().put("uml", new XMIResourceFactoryImpl());
		
		Resource compositionMetamodel = VirtualModelUtil.loadResource(this, virtualModelProperties.getProperty("compositionMetamodel"));
		List<Resource> contributingMetamodels = new ArrayList<Resource>();
		for(String metamodelPath : virtualModelProperties.getProperty("contributingMetamodels").split(",")) {
			if(metamodelPath.startsWith("uri")){
				contributingMetamodels.add(EPackage.Registry.INSTANCE.getEPackage(metamodelPath.substring(4)).eResource());
			} else {
				contributingMetamodels.add(VirtualModelUtil.loadResource(this, metamodelPath));
			}
		}
		this.metamodelManager = new MetamodelManager(contributingMetamodels, compositionMetamodel, this);
		
		String concreteResourcePaths[] = virtualModelProperties.getProperty("contributingModels").split(",");
		this.contributingModels = new ArrayList<Resource>();
		for(int i = 0 ; i < concreteResourcePaths.length ; i++) {
			this.contributingModels.add(VirtualModelUtil.loadResource(this, concreteResourcePaths[i]));
		}
		
		this.vLinkManager = new VirtualLinkManager(virtualModelProperties.getProperty("correspondenceModel"), this);
		vLinkManager.initialize();

		EList<EObject>[] subLists = new EList[this.contributingModels.size()];
		for(int i = 0 ; i < this.contributingModels.size() ; i++) {
			// INFO: we suppose that getContents returns always the same list...
			subLists[i] = this.contributingModels.get(i).getContents();
		}
		
		this.virtualContents = new VirtualContents<EObject>(this, subLists);	
	}
	
	
	public EList<EObject> getContents(){
		return virtualContents;
	}
	
	public EObject translateToVirtualElement(EObject e) {
		if(e == null) {
			return null;
		} else {
			return vLinkManager.getVirtualElement(e);
		}
	}
	
	@Override
	protected void doSave(OutputStream outputStream, Map<?, ?> options)	throws IOException {
		//TODO: save virtual links
		for(Resource r : this.contributingModels) {
			r.save(options);
		}
		vLinkManager.save();
		virtualModelProperties.store(outputStream, null);
	}
	
	@Override
	protected void doUnload() {
		for (Resource r : this.contributingModels) {
			r.unload();
		}
		vLinkManager.getCorrespondenceModel().unload();
	}

	public List<Resource> getContributingModels(){
		return contributingModels;
	}
	
	public VirtualLinkManager getVirtualLinkManager(){
		return this.vLinkManager;
	}
	
	public MetamodelManager getMetamodelManager(){
		return metamodelManager;
	}

	public ResourceSet getVirtualResourceSet() {
		return virtualResourceSet;
	}
	
}